
/*****************************************************************************/
/*                                                                           */
/*  THE HSEVAL HIGH SCHOOL TIMETABLE EVALUATOR                               */
/*  COPYRIGHT (C) 2009, Jeffrey H. Kingston                                  */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         cgi.c                                                      */
/*  MODULE:       Reading CGI data                                           */
/*                                                                           */
/*****************************************************************************/
#include "howard_a.h"
#include "howard_n.h"
#include "cgi.h"
#include "html.h"
#include "url.h"
#include <sys/stat.h>

#define DEBUG1 0

/*****************************************************************************/
/*                                                                           */
/*  CGI - CGI input file, which could be in either Post or Get format        */
/*                                                                           */
/*****************************************************************************/
#define MAX_LINE 200

struct cgi_rec
{
  bool post;
  char *query_string;
  char null_query_string[1];
  char mime_sep[MAX_LINE];
  FILE *rerun;
};


/*****************************************************************************/
/*                                                                           */
/*  void CGIProblem(char *text)                                              */
/*                                                                           */
/*  Report a CGI format problem and exit.                                    */
/*                                                                           */
/*****************************************************************************/

static void CGIProblem(char *text)
{
  HTML html;

  html = HTMLPageBegin("./hseval.cgi", "HSEval home page", stdout);
  HTMLHeadBegin(html);
  HTMLTitleBegin(html);
  HTMLText(html, text);
  HTMLTitleEnd(html);
  HTMLHeadEnd(html);
  HTMLBodyBegin(html);

  HTMLBigHeading(html, "HSEval: format problem in uploaded CGI file");

  HTMLParagraphBegin(html);
  HTMLText(html, "The uploaded CGI file uploaded has at least one error:");
  HTMLParagraphEnd(html);

  HTMLParagraphBegin(html);
  HTMLColouredBoxBegin(html, LightRed);
  HTMLLiteralText(html, text);
  HTMLColouredBoxEnd(html);
  HTMLParagraphEnd(html);

  HTMLParagraphBegin(html);
  HTMLText(html, "Return to the ");
  HTMLJumpFront(html);
  HTMLText(html, ".");
  HTMLParagraphEnd(html);

  HTMLBodyEnd(html);
  HTMLPageEnd(html);
  exit(0);
}


/*****************************************************************************/
/*                                                                           */
/*  CGI CgiBegin(bool rerun)                                                 */
/*                                                                           */
/*  Make a CGI object with an optional rerun file (which will eventually     */
/*  hold a copy of the CGI input).                                           */
/*                                                                           */
/*****************************************************************************/

CGI CgiBegin(bool rerun)
{
  CGI res;  int ch;
  if( DEBUG1 )
    fprintf(stderr, "[ CgiBegin(rerun %s)\n", rerun ? "true" : "false");
  res = (CGI) malloc(sizeof(struct cgi_rec));
  ch = getc(stdin);
  res->null_query_string[0] = '\0';
  if( ch != EOF )
  {
    ungetc(ch, stdin);
    res->post = true;
    res->query_string = NULL;
    if( DEBUG1 )
      fprintf(stderr, "  post\n");
  }
  else
  {
    res->post = false;
    res->query_string = getenv("QUERY_STRING");
    if( res->query_string == NULL )
      res->query_string = res->null_query_string;
    if( DEBUG1 )
      fprintf(stderr, "  get\n");
  }
  res->mime_sep[0] = '\0';
  if( rerun )
  {
    res->rerun = fopen("./rerun", "w");
    if( res->rerun == NULL )
      CGIProblem("cannot open debug file \"./rerun\"");
    fprintf(res->rerun, "rm -f core*\n");
    fprintf(res->rerun, "MALLOC_CHECK_=2 exec ./hseval.cgi <<EOF\n");
    chmod("./rerun", S_IRUSR | S_IWUSR | S_IXUSR | S_IROTH | S_IWOTH | S_IXOTH);
  }
  else
    res->rerun = NULL;
  if( DEBUG1 )
    fprintf(stderr, "] CgiBegin\n");
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void CgiEnd(CGI cgi)                                                     */
/*                                                                           */
/*  End the processing of CGI.                                               */
/*                                                                           */
/*****************************************************************************/

void CgiEnd(CGI cgi)
{
  int ch;
  if( cgi->rerun != NULL )
  {
    if( cgi->query_string == NULL )
      while( (ch = getc(stdin)) != EOF )
	putc(ch, cgi->rerun);
    fprintf(cgi->rerun, "EOF\n");
    fclose(cgi->rerun);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  FILE *CgiRerunFile(CGI cgi)                                              */
/*                                                                           */
/*  Return the rerun file of cgi, or NULL if not requested.                  */
/*                                                                           */
/*****************************************************************************/

FILE *CgiRerunFile(CGI cgi)
{
  return cgi->rerun;
}


/*****************************************************************************/
/*                                                                           */
/*  bool str_begins_with(char *str, char *pat)                               */
/*                                                                           */
/*  true if str begins with pat.                                             */
/*                                                                           */
/*****************************************************************************/

static bool str_begins_with(char *str, char *pat)
{
  return strstr(str, pat) == str;
}


/*****************************************************************************/
/*                                                                           */
/*  char *CgiNextOp(CGI cgi)                                                 */
/*                                                                           */
/*  Assuming we're ready for it, return the next operator, or NULL if none.  */
/*  A non-NULL result may have any length and will lie in malloced memory.   */
/*                                                                           */
/*****************************************************************************/

char *CgiNextOp(CGI cgi)
{
  char line[MAX_LINE], *name, *name_end, *p, *q, *res, last;

  if( cgi->post )
  {
    /* read one mime separator into sep, exiting if not present */
    p = fgets(line, MAX_LINE, stdin);  /* p unused, keep compiler happy */
    if( cgi->rerun != NULL )
      fputs(line, cgi->rerun);
    if( line[0] == '\n' || line[0] == '\0' )
    {
      if( DEBUG1 )
	fprintf(stderr, "CgiNextOp returning NULL");
      return NULL;
    }
    last = line[strlen(line) - 1];
    if( last != '\n' && last != '\r' )
      CGIProblem("missing newline at end of first mime separator");

    /* either save the mime separator or make sure it hasn't changed */
    if( cgi->mime_sep[0] == '\0' )
      strcpy(cgi->mime_sep, line);
    else if( strcmp(line, cgi->mime_sep) != 0 )
      CGIProblem("mime separator changed or not found where expected");

    /* read header lines */
    res = NULL;
    p = fgets(line, MAX_LINE, stdin);  /* p unused, keep compiler happy */
    if( cgi->rerun != NULL )
      fputs(line, cgi->rerun);
    while( line[0] != '\n' && line[0] != '\r' )
    {
      if( str_begins_with(line, "Content-Disposition:") &&
	  (name = strstr(line, "name=\"")) != NULL )
      {
	if( res != NULL )
	  CGIProblem("redundant Content-Disposition line in form-data");
	name += 6;
	if( (name_end = strstr(name, "\"")) == NULL || name_end == name + 1 )
	  CGIProblem("format error on Content-Disposition line in form-data");
	res = malloc(sizeof(char) * (name_end - name + 1));
	for( p = name, q = res;  p != name_end;  p++, q++ )
	  *q = *p;
	*q = '\0';
      }
      p = fgets(line, MAX_LINE, stdin);  /* p unused, keep compiler happy */
      if( cgi->rerun != NULL )
	fputs(line, cgi->rerun);
    }
  }
  else
  {
    /* read the key into op */
    res = URLDecodePartial(cgi->query_string, "=&", &cgi->query_string);
    if( *cgi->query_string == '&' )
      CGIProblem("unexpected & in CGI string");
    if( *cgi->query_string != '\0' )
      cgi->query_string++;
  }
  if( DEBUG1 )
    fprintf(stderr, "CgiNextOp returning %s\n", res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *CgiNextOneLineVal(CGI cgi)                                         */
/*                                                                           */
/*  Assuming that the file is positioned at the start of a value, and that   */
/*  the value is a simple one-line string, get that value.  Strip any        */
/*  trailing newline etc. characters.                                        */
/*                                                                           */
/*  A non-NULL result may have any length and will lie in malloced memory.   */
/*                                                                           */
/*****************************************************************************/

char *CgiNextOneLineVal(CGI cgi)
{
  char line[MAX_LINE], *p, *res;
  if( cgi->post )
  {
    /* read the next line from stdin and make that the value */
    p = fgets(line, MAX_LINE, stdin);  /* p unused; keep compiler happy */
    if( cgi->rerun != NULL )
      fputs(line, cgi->rerun);
    for( p = &line[strlen(line)-1];  p >= line && (*p=='\n' || *p=='\r');  p-- )
      *p = '\0';
    res = malloc(sizeof(char) * (p - line + 1));
    strcpy(res, line);
  }
  else
  {
    /* read the value from the query string */
    res = URLDecodePartial(cgi->query_string, "=&", &cgi->query_string);
    if( *cgi->query_string == '=' )
      CGIProblem("unexpected = in CGI string");
    if( *cgi->query_string != '\0' )
      cgi->query_string++;
  }
  if( DEBUG1 )
    fprintf(stderr, "CgiNextOneLineVal returning %s\n", res);
  return res;
}
