
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_task_set.c                                             */
/*  DESCRIPTION:  Task sets                                                  */
/*                                                                           */
/*****************************************************************************/
#include "khe_interns.h"

/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK_SET                                                             */
/*                                                                           */
/*****************************************************************************/

struct khe_task_set_rec
{
  KHE_SOLN			soln;			/* encl solution     */
  ARRAY_KHE_TASK		tasks;			/* soln tasks        */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK_SET KheTaskSetMake(KHE_SOLN soln)                               */
/*                                                                           */
/*  Make a new task set with these attributes.                               */
/*                                                                           */
/*****************************************************************************/

KHE_TASK_SET KheTaskSetMake(KHE_SOLN soln)
{
  KHE_TASK_SET res;  HA_ARENA a;
  res = KheSolnGetTaskSetFromFreeList(soln);
  if( res == NULL )
  {
    /* MMakeCount(res, 50); */
    a = KheSolnArena(soln);
    HaMake(res, a);
    HaArrayInit(res->tasks, a);
  }
  else
    HaArrayClear(res->tasks);
  res->soln = soln;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetDelete(KHE_TASK_SET ts)                                   */
/*                                                                           */
/*  Delete ts, that is, return it to its solution's free list.               */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetDelete(KHE_TASK_SET ts)
{
  KheSolnAddTaskSetToFreeList(ts->soln, ts);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetFree(KHE_TASK_SET ts)                                     */
/*                                                                           */
/*  Free ts.                                                                 */
/*                                                                           */
/*****************************************************************************/

/* ***
void KheTaskSetFree(KHE_TASK_SET ts)
{
  MArrayFree(ts->tasks);
  MFree(ts);
}
*** */


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetClear(KHE_TASK_SET ts)                                    */
/*                                                                           */
/*  Clear ts back to the empty set of tasks.                                 */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetClear(KHE_TASK_SET ts)
{
  HaArrayClear(ts->tasks);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetDropFromEnd(KHE_TASK_SET ts, int n)                       */
/*                                                                           */
/*  Remove the last n tasks from ts.                                         */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetDropFromEnd(KHE_TASK_SET ts, int n)
{
  HaArrayDeleteLastSlice(ts->tasks, n);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetAddTask(KHE_TASK_SET ts, KHE_TASK task)                   */
/*                                                                           */
/*  Add task to ts.                                                          */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetAddTask(KHE_TASK_SET ts, KHE_TASK task)
{
  HaArrayAddLast(ts->tasks, task);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetDeleteTask(KHE_TASK_SET ts, KHE_TASK task)                */
/*                                                                           */
/*  Delete task from ts.                                                     */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetDeleteTask(KHE_TASK_SET ts, KHE_TASK task)
{
  int pos;
  if( !HaArrayContains(ts->tasks, task, &pos) )
    HnAbort("KheTaskSetDeleteTask:  ts does not contain task");
  HaArrayDeleteAndPlug(ts->tasks, pos);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTaskSetContainsTask(KHE_TASK_SET ts, KHE_TASK task, int *pos)    */
/*                                                                           */
/*  Return true and set *pos if ts contains task.                            */
/*                                                                           */
/*****************************************************************************/

bool KheTaskSetContainsTask(KHE_TASK_SET ts, KHE_TASK task, int *pos)
{
  return HaArrayContains(ts->tasks, task, pos);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTaskSetTaskCount(KHE_TASK_SET ts)                                 */
/*                                                                           */
/*  Return the number of tasks in ts.                                        */
/*                                                                           */
/*****************************************************************************/

int KheTaskSetTaskCount(KHE_TASK_SET ts)
{
  return HaArrayCount(ts->tasks);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_TASK KheTaskSetTask(KHE_TASK_SET ts, int i)                          */
/*                                                                           */
/*  Return the i'th task of ts.                                              */
/*                                                                           */
/*****************************************************************************/

KHE_TASK KheTaskSetTask(KHE_TASK_SET ts, int i)
{
  return HaArray(ts->tasks, i);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetSort(KHE_TASK_SET ts,                                     */
/*    int(*compar)(const void *, const void *))                              */
/*                                                                           */
/*  Sort the tasks of ts.                                                    */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetSort(KHE_TASK_SET ts, int(*compar)(const void *, const void *))
{
  HaArraySort(ts->tasks, compar);
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetSortUnique(KHE_TASK_SET ts,                               */
/*    int(*compar)(const void *, const void *))                              */
/*                                                                           */
/*  Sort and uniquefy the tasks of ts.                                       */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetSortUnique(KHE_TASK_SET ts,
  int(*compar)(const void *, const void *))
{
  HaArraySortUnique(ts->tasks, compar);
}


/*****************************************************************************/
/*                                                                           */
/*  int KheTaskSetTotalDuration(KHE_TASK_SET ts)                             */
/*                                                                           */
/*  Return the total duration of the tasks of ts.                            */
/*                                                                           */
/*****************************************************************************/

int KheTaskSetTotalDuration(KHE_TASK_SET ts)
{
  int res, i;  KHE_TASK task;
  res = 0;
  HaArrayForEach(ts->tasks, task, i)
    res += KheTaskTotalDuration(task);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  float KheTaskSetTotalWorkload(KHE_TASK_SET ts)                           */
/*                                                                           */
/*  Return the total workload of the tasks of ts.                            */
/*                                                                           */
/*****************************************************************************/

float KheTaskSetTotalWorkload(KHE_TASK_SET ts)
{
  float res;  int i;  KHE_TASK task;
  res = 0.0;
  HaArrayForEach(ts->tasks, task, i)
    res += KheTaskTotalWorkload(task);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskSetUnGroup(KHE_TASK_SET ts)                                  */
/*                                                                           */
/*  Ungroup the tasks of ts.  That is, unassign them from their              */
/*  immediate parents, but then assign them to the resources that            */
/*  those parents are assigned to.                                           */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetUnGroup(KHE_TASK_SET ts)
{
  KHE_TASK task;  int i;  KHE_RESOURCE r;
  HaArrayForEach(ts->tasks, task, i)
  {
    r = KheTaskAsstResource(task);
    if( KheTaskUnAssign(task) && r != NULL )
      KheTaskAssignResource(task, r);
  }
}


/*****************************************************************************/
/*                                                                           */
/* void KheTaskSetDebug(KHE_TASK_SET ts, int verbosity, int indent, FILE *fp)*/
/*                                                                           */
/*  Debug print of task set ts onto fp with the given verbosity and indent.  */
/*                                                                           */
/*****************************************************************************/

void KheTaskSetDebug(KHE_TASK_SET ts, int verbosity, int indent, FILE *fp)
{
  KHE_TASK task;  int i;
  if( verbosity >= 1 )
  {
    if( indent >= 0 )
      fprintf(fp, "%*s", indent, "");
    fprintf(fp, "{");
    HaArrayForEach(ts->tasks, task, i)
    {
      if( i > 0 )
	fprintf(fp, ", ");
      KheTaskDebug(task, verbosity, -1, fp);
    }
    fprintf(fp, "}");
    if( indent >= 0 )
      fprintf(fp, "\n");
  }
}
