
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU Generar Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sr_time_sweep.c                                        */
/*  DESCRIPTION:  Time sweep resource assignment                             */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"

#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0
#define DEBUG4 0
#define DEBUG5 0
#define DEBUG6 0
#define DEBUG7 0

typedef HA_ARRAY(KHE_RESOURCE_MATCHING_DEMAND_SET)
  ARRAY_KHE_RESOURCE_MATCHING_DEMAND_SET;

/*****************************************************************************/
/*                                                                           */
/*  KHE_MONITOR_FRAME                                                        */
/*                                                                           */
/*  A monitor frame is a partition of the monitors which need cutoffs,       */
/*  so that all the monitors whose initial cutoff time lies in a given       */
/*  part of the frame are grouped together under the index of that part.     */
/*                                                                           */
/*****************************************************************************/

typedef HA_ARRAY(KHE_MONITOR) ARRAY_KHE_MONITOR;

typedef struct khe_monitor_frame_part_rec {
  ARRAY_KHE_MONITOR	monitors;
} *KHE_MONITOR_FRAME_PART;

typedef HA_ARRAY(KHE_MONITOR_FRAME_PART) ARRAY_KHE_MONITOR_FRAME_PART;

typedef struct khe_monitor_frame_rec {
  KHE_FRAME			frame;
  ARRAY_KHE_MONITOR		active_monitors;
  ARRAY_KHE_MONITOR_FRAME_PART	parts;
} *KHE_MONITOR_FRAME;


/*****************************************************************************/
/*                                                                           */
/*  Submodule "monitor frames (private)"                                     */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_MONITOR_FRAME KheMonitorFrameMake(KHE_FRAME frame)                   */
/*                                                                           */
/*  Make a new, empty monitor frame suitable for use with frame.             */
/*                                                                           */
/*****************************************************************************/

static KHE_MONITOR_FRAME KheMonitorFrameMake(KHE_FRAME frame, HA_ARENA a)
{
  KHE_MONITOR_FRAME res;  KHE_MONITOR_FRAME_PART part;  int i;
  HaMake(res, a);
  res->frame = frame;
  HaArrayInit(res->active_monitors, a);
  HaArrayInit(res->parts, a);
  for( i = 0;  i < KheFrameTimeGroupCount(frame);  i++ )
  {
    HaMake(part, a);
    HaArrayInit(part->monitors, a);
    HaArrayAddLast(res->parts, part);
  }
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFrameAddMonitor(KHE_MONITOR_FRAME mf,                     */
/*    KHE_MONITOR m, KHE_TIME initial_cutoff_time)                           */
/*                                                                           */
/*  Add m to mf at the position where initial_cutoff_time occurs in          */
/*  frame.  Or if initial_cutoff_time is NULL, do nothing.                   */
/*                                                                           */
/*****************************************************************************/

static void KheMonitorFrameAddMonitor(KHE_MONITOR_FRAME mf,
  KHE_MONITOR m, KHE_TIME initial_cutoff_time)
{
  int index;  KHE_MONITOR_FRAME_PART part;
  if( initial_cutoff_time != NULL )
  {
    index = KheFrameTimeIndex(mf->frame, initial_cutoff_time);
    part = HaArray(mf->parts, index);
    HaArrayAddLast(part->monitors, m);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFrameAddAndCutOffAllMonitors(KHE_MONITOR_FRAME mf)        */
/*                                                                           */
/*  Add all relevant monitors to mf, and cut them all off at index 0.        */
/*                                                                           */
/*****************************************************************************/

static void KheMonitorFrameAddAndCutOffAllMonitors(KHE_MONITOR_FRAME mf,
  KHE_RESOURCE_GROUP rg, KHE_SOLN soln)
{
  KHE_TIME t;  int i, j;  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;
  KHE_RESOURCE r;  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;  KHE_MONITOR m;
  if( DEBUG6 )
    fprintf(stderr, "[ KheMonitorFrameAddAndCutOffAllMonitors(...):\n");
  for( i = 0;  i < KheResourceGroupResourceCount(rg);  i++ )
  {
    r = KheResourceGroupResource(rg, i);
    for( j = 0;  j < KheSolnResourceMonitorCount(soln, r);  j++ )
    {
      m = KheSolnResourceMonitor(soln, r, j);
      if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
      {
	laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
	KheLimitActiveIntervalsMonitorSetCutoffIndex(laim, 0);
        t = KheLimitActiveIntervalsMonitorInitialCutoffTime(laim);
	KheMonitorFrameAddMonitor(mf, m, t);
	if( DEBUG6 )
	  KheMonitorDebug(m, 2, 2, stderr);
      }
      else if( KheMonitorTag(m) == KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG )
      {
	cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) m;
	KheClusterBusyTimesMonitorSetCutoffIndex(cbtm, 0);
        t = KheClusterBusyTimesMonitorInitialCutoffTime(cbtm);
	KheMonitorFrameAddMonitor(mf, m, t);
	if( DEBUG6 )
	  KheMonitorDebug(m, 2, 2, stderr);
      }
    }
  }
  if( DEBUG6 )
    fprintf(stderr, "] KheMonitorFrameAddAndCutOffAllMonitors returning\n");
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFrameUpdateMonitorCutoffs(KHE_MONITOR_FRAME mf,           */
/*    int frame_index)                                                       */
/*                                                                           */
/*  We are now starting on position frame_index of the frame, so add         */
/*  in the monitors that start there, cut off all monitors at the last       */
/*  time of that frame part, and remove from the current list all            */
/*  monitors that don't need cutting off after here.                         */
/*                                                                           */
/*****************************************************************************/

static void KheMonitorFrameUpdateMonitorCutoffs(KHE_MONITOR_FRAME mf,
  int frame_index)
{
  KHE_MONITOR_FRAME_PART part;  int i, tg_count;  KHE_TIME_GROUP tg;
  KHE_POLARITY po;  KHE_TIME t;  bool still_active;
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;  KHE_MONITOR m;
  int initial, adding, dropping;

  /*  add in the monitors that start at frame_index */
  if( DEBUG6 )
    fprintf(stderr, "[ KheMonitorFrameUpdateMonitorCutoffs(mf, %d):\n",
      frame_index);
  part = HaArray(mf->parts, frame_index);
  initial = HaArrayCount(mf->active_monitors);
  adding = HaArrayCount(part->monitors);
  HaArrayAppend(mf->active_monitors, part->monitors, i);

  /* find the time we need to be cutting off at */
  tg = KheFrameTimeGroup(mf->frame, frame_index, &po);
  tg_count = KheTimeGroupTimeCount(tg);
  HnAssert(tg_count > 0, "KheTimeSweepAssignResources: empty part");
  t = KheTimeGroupTime(tg, tg_count - 1);

  /* cut off all active monitors, simultaneously removing from the active */
  /* list those that don't need to be cut off again */
  dropping = 0;
  for( i = 0;  i < HaArrayCount(mf->active_monitors);  i++ )
  {
    m = HaArray(mf->active_monitors, i);
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
    {
      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      still_active = KheLimitActiveIntervalsMonitorSetCutoffTime(laim, t);
    }
    else if( KheMonitorTag(m) == KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG )
    {
      cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) m;
      still_active = KheClusterBusyTimesMonitorSetCutoffTime(cbtm, t);
    }
    else
    {
      HnAbort("KheMonitorFrameUpdateMonitorCutoffs internal error: tag");
      still_active = false;  /* keep compiler happy */
    }
    if( !still_active )
    {
      HaArrayDeleteAndPlug(mf->active_monitors, i);
      i--;
      dropping++;
    }
  }
  if( DEBUG6 )
  {
    HaArrayForEach(mf->active_monitors, m, i)
      KheMonitorDebug(m, 2, 2, stderr);
    fprintf(stderr, "  %d + %d = %d - %d = %d:\n", initial,
      adding, initial + adding, dropping, HaArrayCount(mf->active_monitors));
    fprintf(stderr, "] KheMonitorFrameUpdateMonitorCutoffs returning\n");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFrameRemoveMonitorCutoffs(KHE_MONITOR_FRAME mf)           */
/*                                                                           */
/*  Remove cutoffs from the remaining active monitors.                       */
/*                                                                           */
/*****************************************************************************/

static void KheMonitorFrameRemoveMonitorCutoffs(KHE_MONITOR_FRAME mf)
{
  KHE_LIMIT_ACTIVE_INTERVALS_MONITOR laim;  KHE_MONITOR m;
  KHE_CLUSTER_BUSY_TIMES_MONITOR cbtm;  int i;
  HaArrayForEach(mf->active_monitors, m, i)
  {
    if( KheMonitorTag(m) == KHE_LIMIT_ACTIVE_INTERVALS_MONITOR_TAG )
    {
      laim = (KHE_LIMIT_ACTIVE_INTERVALS_MONITOR) m;
      KheLimitActiveIntervalsMonitorSetCutoffTime(laim, NULL);
    }
    else if( KheMonitorTag(m) == KHE_CLUSTER_BUSY_TIMES_MONITOR_TAG )
    {
      cbtm = (KHE_CLUSTER_BUSY_TIMES_MONITOR) m;
      KheClusterBusyTimesMonitorSetCutoffTime(cbtm, NULL);
    }
    else
      HnAbort("KheMonitorFrameRemoveMonitorCutoffs internal error: tag");
  }
}


/*****************************************************************************/
/*                                                                           */
/*  void KheMonitorFrameDebug(KHE_MONITOR_FRAME mf, int indent, FILE *fp)    */
/*                                                                           */
/*  Debug print of mf onto fp with the given indent.                         */
/*                                                                           */
/*****************************************************************************/

static void KheMonitorFrameDebug(KHE_MONITOR_FRAME mf, int indent, FILE *fp)
{
  int i, j;  KHE_TIME_GROUP tg;  KHE_POLARITY po;
  KHE_MONITOR_FRAME_PART part;  KHE_MONITOR m;
  fprintf(fp, "%*s[ Monitor frame (%d parts):\n", indent, "",
    KheFrameTimeGroupCount(mf->frame));
  for( i = 0;  i < KheFrameTimeGroupCount(mf->frame);  i++ )
  {
    tg = KheFrameTimeGroup(mf->frame, i, &po);
    part = HaArray(mf->parts, i);
    fprintf(fp, "%*s[ ", indent + 2, "");
    KheTimeGroupDebug(tg, 1, 0, fp);
    HaArrayForEach(part->monitors, m, j)
      KheMonitorDebug(m, 2, indent + 4, fp);
    fprintf(fp, "%*s]\n", indent + 2, "");
  }
  fprintf(fp, "%*s]\n", indent, "");
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "main function"                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  int KheDefaultRematchMaxGroups(KHE_FRAME frame)                          */
/*                                                                           */
/*  Return a default maximum number of rematch time groups.  The larger      */
/*  the frame size, the smaller the number of time groups, for speed.        */
/*                                                                           */
/*****************************************************************************/

static int KheDefaultRematchMaxGroups(KHE_FRAME frame)
{
  int count;
  count = KheFrameTimeGroupCount(frame);
  if( count <= 4 * 7 )
    return 7;
  else if( count <= 26 * 7 )
    return 4;
  else
    return 1;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheSaveDemandSets(KHE_FRAME frame)                                  */
/*                                                                           */
/*  Return true if we are saving demand sets.  This is only feasible on      */
/*  fairly short frames.                                                     */
/*                                                                           */
/*****************************************************************************/

static bool KheSaveDemandSets(KHE_FRAME frame)
{
  return KheFrameTimeGroupCount(frame) <= 4 * 7;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheTimeSweepAssignResources(KHE_SOLN soln, KHE_RESOURCE_GROUP rg,   */
/*    KHE_OPTIONS options)                                                   */
/*                                                                           */
/*  Carry out time sweep assignment of the resources of rg in soln.          */
/*                                                                           */
/*****************************************************************************/
#define min(a, b) ((a) < (b) ? (a) : (b))

bool KheTimeSweepAssignResources(KHE_SOLN soln, KHE_RESOURCE_GROUP rg,
  KHE_OPTIONS options)
{
  KHE_RESOURCE_MATCHING_SOLVER rms;  KHE_FRAME frame;  KHE_POLARITY po;
  int frame_pos, frame_len, rematch_max_groups, count, parts, i, redo_count;
  KHE_COST init_cost, global_init_cost;  KHE_TIME_GROUP tg;
  bool matching_off, edge_adjust_off, preserve_existing;
  bool cutoff_off, redo_off, rematch_off, rematch_redo;
  ARRAY_KHE_RESOURCE_MATCHING_DEMAND_SET demand_sets;
  KHE_RESOURCE_MATCHING_DEMAND_SET demand_set;
  KHE_STATS_TIMER st_all, st_cutoffs, st_matches;
  float time_all, time_cutoffs, time_matches;
  KHE_MONITOR_FRAME mf;  HA_ARENA a;

  if( DEBUG7 )
  {
    fprintf(stderr, "at start of KheTimeSweepAssignResources, arena set is:\n");
    HaArenaSetDebug(KheSolnArenaSet(soln), 1, 2, stderr);
  }
  a = KheSolnArenaBegin(soln);
  if( DEBUG3 )
  {
    /* time tests */
    st_all = KheStatsTimerMake(a);
    st_cutoffs = KheStatsTimerMake(a);
    st_matches = KheStatsTimerMake(a);
    time_all = time_cutoffs = time_matches = 0.0;
  }

  /* get the solver and frame */
  rms = KheResourceMatchingSolverMake(soln, rg, a);
  frame = KheOptionsFrame(options, "gs_common_frame", soln);
  frame_len = KheFrameTimeGroupCount(frame);
  global_init_cost = KheSolnCost(soln);

  /* get the options */
  matching_off = KheOptionsGetBool(options,
    "rs_time_sweep_matching_off", false);
  edge_adjust_off = KheOptionsGetBool(options,
    "rs_time_sweep_edge_adjust_off", false);
  preserve_existing = ! KheOptionsGetBool(options,
    "rs_time_sweep_preserve_existing_off", false);
  cutoff_off = KheOptionsGetBool(options,
    "rs_time_sweep_cutoff_off", false);
  redo_off = KheOptionsGetBool(options,
    "rs_time_sweep_redo_off", false);
  rematch_off = KheOptionsGetBool(options,
    "rs_time_sweep_rematch_off", false);
  rematch_max_groups = KheOptionsGetInt(options,
    "rs_time_sweep_rematch_max_groups", KheDefaultRematchMaxGroups(frame));
  rematch_redo = false;

  if( DEBUG1 )
  {
    fprintf(stderr, "[ KheTimeSweepAssignResources(soln, rg, options)\n");
    fprintf(stderr, "  matching_off = %s\n", matching_off ? "true" : "false");
  }

  /* get the demand sets, one for each time group of frame; or if the */
  /* frame is too long to make that safe, get them one by one later   */
  HaArrayInit(demand_sets, a);
  if( KheSaveDemandSets(frame) )
  {
    for( frame_pos = 0;  frame_pos < frame_len;  frame_pos++ )
    {
      tg = KheFrameTimeGroup(frame, frame_pos, &po);
      demand_set = KheResourceMatchingDemandSetMake(rms, preserve_existing);
      KheResourceMatchingDemandSetAddTimeGroup(demand_set, tg);
      HaArrayAddLast(demand_sets, demand_set);
    }
  }

  if( !cutoff_off )
  {
    mf = KheMonitorFrameMake(frame, a);
    KheMonitorFrameAddAndCutOffAllMonitors(mf, rg, soln);
    if( DEBUG5 )
      KheMonitorFrameDebug(mf, 2, stderr);
  }
  else
    mf = NULL;  /* keep compiler happy */

  /* do one solve for each time group of frame */
  for( frame_pos = 0;  frame_pos < frame_len;  frame_pos++ )
  {
    /* get time group and demand set for frame_pos */
    tg = KheFrameTimeGroup(frame, frame_pos, &po);
    if( KheSaveDemandSets(frame) )
      demand_set = HaArray(demand_sets, frame_pos);
    else
    {
      demand_set = KheResourceMatchingDemandSetMake(rms, preserve_existing);
      KheResourceMatchingDemandSetAddTimeGroup(demand_set, tg);
    }

    /* install cutoffs, unless cutoff_off */
    if( !cutoff_off )
    {
      if( DEBUG3 )
	KheStatsTimerReset(st_cutoffs);
      KheMonitorFrameUpdateMonitorCutoffs(mf, frame_pos);
      if( DEBUG3 )
	time_cutoffs += KheStatsTimerNow(st_cutoffs);
    }

    /* solve tg */
    if( DEBUG2 )
      fprintf(stderr, "  calling KheResourceMatchingSolverSolve from"
	" KheTimeSweepAssignResources (a): %s\n", KheTimeGroupId(tg));
    if( DEBUG3 )
      KheStatsTimerReset(st_matches);
    KheResourceMatchingSolverSolve(rms, demand_set, matching_off,
      edge_adjust_off, options);
    if( !KheSaveDemandSets(frame) )
      KheResourceMatchingDemandSetClear(demand_set);
    if( DEBUG3 )
      time_matches += KheStatsTimerNow(st_matches);

    /* rematch, unless rematch_off */
    if( !rematch_off && !KheOptionsTimeLimitReached(options) )
    {
      count = min(rematch_max_groups, frame_pos + 1);
      do
      {
	if( DEBUG4 )
	  fprintf(stderr, "  and rematching 2 .. %d parts\n", count);
	init_cost = KheSolnCost(soln);
	demand_set = KheResourceMatchingDemandSetMake(rms, false);
	for( parts = 2;  parts <= count;  parts++ )
	{
	  for( i = 0;  i < parts;  i++ )
	  {
	    tg = KheFrameTimeGroup(frame, frame_pos - i, &po);
	    KheResourceMatchingDemandSetAddTimeGroup(demand_set, tg);
	  }
	  if( DEBUG2 )
	  {
	    fprintf(stderr, "  calling KheResourceMatchingSolverSolve from"
	      " KheTimeSweepAssignResources (b):");
	    for( i = 0;  i < parts;  i++ )
	    {
	      tg = KheFrameTimeGroup(frame, frame_pos - i, &po);
	      fprintf(stderr, " %s", KheTimeGroupId(tg));
	    }
	    fprintf(stderr, "\n");
	  }
	  if( DEBUG3 )
	    KheStatsTimerReset(st_matches);
	  KheResourceMatchingSolverSolve(rms, demand_set, matching_off,
	    edge_adjust_off, options);
          KheResourceMatchingDemandSetClear(demand_set);
	  if( DEBUG3 )
	    time_matches += KheStatsTimerNow(st_matches);
	}
	KheResourceMatchingDemandSetDelete(demand_set);
      } while( rematch_redo && KheSolnCost(soln) < init_cost );
    }
  }

  /* remove cutoffs, unless cutoff_off */
  if( !cutoff_off )
  {
    if( DEBUG3 )
      KheStatsTimerReset(st_cutoffs);
    KheMonitorFrameRemoveMonitorCutoffs(mf);
    if( DEBUG3 )
      time_cutoffs += KheStatsTimerNow(st_cutoffs);
  }

  /* if time permits, redo while improving (no cutoffs, no rematch) */
  /* although redo no more than twice, because of diminishing returns */
  if( !redo_off && !KheOptionsTimeLimitReached(options) )
  {
    redo_count = 0;
    do
    {
      init_cost = KheSolnCost(soln);
      for( frame_pos = 0;  frame_pos < frame_len;  frame_pos++ )
      {
	/* get time group and demand set for frame_pos */
	tg = KheFrameTimeGroup(frame, frame_pos, &po);
	if( KheSaveDemandSets(frame) )
	  demand_set = HaArray(demand_sets, frame_pos);
	else
	{
	  demand_set = KheResourceMatchingDemandSetMake(rms, preserve_existing);
	  KheResourceMatchingDemandSetAddTimeGroup(demand_set, tg);
	}

	/* solve tg */
	if( DEBUG2 )
	  fprintf(stderr, "  calling KheResourceMatchingSolverSolve from"
	    " KheTimeSweepAssignResources (c): %s\n", KheTimeGroupId(tg));
	if( DEBUG3 )
	  KheStatsTimerReset(st_matches);
	KheResourceMatchingSolverSolve(rms, demand_set, matching_off,
	  edge_adjust_off, options);
	redo_count++;
	if( !KheSaveDemandSets(frame) )
	  KheResourceMatchingDemandSetClear(demand_set);
	if( DEBUG3 )
	  time_matches += KheStatsTimerNow(st_matches);
      }
    } while( redo_count < 2 && KheSolnCost(soln) < init_cost &&
	!KheOptionsTimeLimitReached(options) );
  }

  /* debug, reclaim all memory, and return */
  if( DEBUG3 )
  {
    time_all += KheStatsTimerNow(st_all);
    fprintf(stderr, "  timing:  all = %.1f, cutoffs = %.1f, matches = %.1f\n",
      time_all, time_cutoffs, time_matches);
  }
  if( DEBUG1 )
    fprintf(stderr, "] KheTimeSweepAssignResources ret. %s (%.5f -> %.5f)\n",
      KheSolnCost(soln) < global_init_cost ? "true" : "false",
      KheCostShow(global_init_cost), KheCostShow(KheSolnCost(soln)));
  KheSolnArenaEnd(soln, a);
  if( DEBUG7 )
  {
    fprintf(stderr, "at end of KheTimeSweepAssignResources, arena set is:\n");
    HaArenaSetDebug(KheSolnArenaSet(soln), 1, 2, stderr);
  }
  return KheSolnCost(soln) < global_init_cost;
}
